#include "tool.h"

float costLimit, attLimit;

float FixedMitigations[TotalMitigations+1];
float Data[RunTotal+1][TotalMitigations+2];
float Distance[RunTotal+1];
float mArray[TotalMitigations+1];

int mCounter,era,run,idleRun,TotalInstances,instanceCounter,displayMode;
float MinCost,MaxCost,MinAtt,MaxAtt;
float infinity,small;

int recentSetMitigation;
float recentSetMitigationStatus;

int main(int argc, char **argv)
{
	int costFlag = 0;
	int attFlag = 0;
	int displayFlag = 0;
	char *costValue = NULL;
	char *attValue = NULL;
	char *displayValue = NULL;
	int c;

	infinity = pow(10,20);
	small = pow(10,-20);
	
	opterr = 0;

	while ((c = getopt(argc, argv, "a:c:d:")) != -1)
	{
		switch (c)
		{
			case 'a':
				attFlag = 1;
				attValue = optarg;
				break;
			case 'c':
				costFlag = 1;
				costValue = optarg;
				break;
			case 'd':
				displayFlag = 1;
				displayValue = optarg;
				break;
			case '?':
				printf("You entered an unknown option. The options must have the format -a AttainmentLowerLimit -c CostUpperLimit -d DisplayMode.\n");
				printf("Display mode 1 prints the final results.\n");
				printf("Display mode 2 prints median and spread of each mitigation-fixing round.\n");
				printf("Display mode 3 is for debugging purposes only.\n");
				break;
			default:
				break;
		}
	}

	if (attFlag == 1 && attValue != NULL)
		attLimit = atof(attValue);
	else
		attLimit = -infinity;

	if (costFlag == 1 && costValue != NULL)
		costLimit = atof(costValue);
	else
		costLimit = infinity;

	if (displayFlag == 1 && displayValue != NULL)
		displayMode = atoi(displayValue);
	else
		displayMode = 1;

	float att, cost;
	srand((unsigned int)time(NULL));

	//set all mitigations to non-fixed value of -1
	for (mCounter = 1; mCounter <= TotalMitigations; mCounter++)
		FixedMitigations[mCounter] = -1;

	if (displayMode == 2)
		printf("Era, Mitigation Number, Mitigation Value, Median of Cost, Spread of Cost, Median of Attainment, Spread of Attainment\n");

	for (era = 1; era <= TotalMitigations; era++)
	{
		TotalInstances = 0;
		MinCost = infinity;
		MaxCost = -infinity;

		MinAtt = infinity;
		MaxAtt = -infinity;

		run = 0;
		idleRun = 0;

		while (run < RunTotal)
		{
			for (mCounter = 1; mCounter <= TotalMitigations; mCounter++)
			{
				//if in the previous runs the mitigation is fixed to a certain value (0 or 1) then use that value. otherwise, select it at random
				if (FixedMitigations[mCounter] == -1)
					mArray[mCounter] = selectValue(0,1);
				else
					mArray[mCounter] = FixedMitigations[mCounter];
			}	
			//find the cost and att using these mitigations
			model(&cost,&att,mArray);

			if (cost < costLimit || att > attLimit || idleRun >= 1000)
			{
				//store the current instance
				addInstance(cost,att);

				//this allows the instances to be added that are desired in the range
				run++;
			}
			else
				idleRun++;
		}

		//find the sweet spot and distances from sweet spot for each distance
		sweetSpot();

		//rank the mitigations and find the mitigation that should be fixed
		rankMitigations();

		if (displayMode == 2)
			reportMedianAndSpread();
	}

	//show the final results
	if (displayMode == 1 || displayMode == 3)
	{
		for (mCounter = 1; mCounter <= TotalMitigations; mCounter++)
			mArray[mCounter] = FixedMitigations[mCounter];
		model(&cost,&att,mArray);
		for (mCounter=1; mCounter<=TotalMitigations; mCounter++)
			printf("m[%d],",mCounter);
		printf("cost,attainment\n");
		for (mCounter=1; mCounter<=TotalMitigations; mCounter++)
			printf("%.0f,", mArray[mCounter]);
		printf("%.1f,%.5f\n",cost,att);
	}
}

void reportMedianAndSpread(void)
{
	float tempCostArray[TotalInstances], tempAttArray[TotalInstances];
	float costMedian,costSpread,attMedian,attSpread;
	
	//sort the cost and att (individually) and find the median and spread
	for (instanceCounter = 1; instanceCounter <= TotalInstances; instanceCounter++)
	{
		tempCostArray[instanceCounter] = Data[instanceCounter][1];
		tempAttArray[instanceCounter] = Data[instanceCounter][2];
	}
	findMedianAndSpread(tempCostArray,TotalInstances,&costMedian,&costSpread);
	findMedianAndSpread(tempAttArray,TotalInstances,&attMedian,&attSpread);

	printf("%d,%d,%.0f,%.5f,%.5f,%.5f,%.5f\n",era,recentSetMitigation,recentSetMitigationStatus,costMedian,costSpread,attMedian,attSpread);
}

void findMedianAndSpread(float inputArray[], int size, float *median, float *spread)
{
	float tempValue;
	int i,j;
	float tempArray[size];

	for (i = 1; i <= size; i++)
		tempArray[i] = inputArray[i];
	
	for (i = 1; i <= size; i++)
	{
		tempValue = tempArray[i];
		j = i;
		
		while ((j > 1) && (tempArray[j-1] > tempValue))
		{
			tempArray[j] = tempArray[j-1];
			j = j - 1;
		}
		tempArray[j] = tempValue;
	}

	*median = tempArray[size/2];
	*spread = tempArray[3*size/4] - tempArray[size/2];
}

float findBestDistance(float inputArray[], int size)
{
	float tempValue;
	int i,j;
	float tempArray[size];

	for (i = 1; i <= size; i++)
		tempArray[i] = inputArray[i];
	
	for (i = 1; i <= size; i++)
	{
		tempValue = tempArray[i];
		j = i;
		
		while ((j > 1) && (tempArray[j-1] > tempValue))
		{
			tempArray[j] = tempArray[j-1];
			j = j - 1;
		}
		tempArray[j] = tempValue;
	}

	return tempArray[Best];
}

void rankMitigations()
{
	float tempScoreOff[TotalInstances],tempScoreOn[TotalInstances],tempBestFreqCount[TotalInstances][2],tempRestFreqCount[TotalInstances][2];
	
	for (mCounter = 1; mCounter <= TotalMitigations; mCounter++)
	{
		tempBestFreqCount[mCounter][0] = 0;
		tempBestFreqCount[mCounter][1] = 0;
		tempRestFreqCount[mCounter][0] = 0;
		tempRestFreqCount[mCounter][1] = 0;
	}
	
	float bestValue = findBestDistance(Distance,TotalInstances);

	for (instanceCounter = 1; instanceCounter <= TotalInstances; instanceCounter++)
	{
		if (displayMode == 3) 
		{
			for (mCounter=1; mCounter<=TotalMitigations; mCounter++)
				printf("%.0f,",Data[instanceCounter][mCounter+2]); 
			printf("%.3f,%.3f,\t", Data[instanceCounter][1],Data[instanceCounter][2]);
		}

		//if it is in the Best distance from the sweet spot, count the frequency of each mitigation (0 and 1) for the best instances
		if (Distance[instanceCounter] <= bestValue)
		{
			if (displayMode == 3) printf("%.3f best from %.3f\n", Distance[instanceCounter],bestValue);
			for (mCounter = 1; mCounter <= TotalMitigations; mCounter++)
			{
				//keep track of the mitigation's counts if it is not already fixed
				if (FixedMitigations[mCounter] == -1)
				{
					if (Data[instanceCounter][mCounter+2] == 0)
						tempBestFreqCount[mCounter][0]++;
					else if (Data[instanceCounter][mCounter+2] == 1)
						tempBestFreqCount[mCounter][1]++;
				}
			}
		}
		//else it is in the Rest distance from the sweet spot and so count the frequency of each mitigation (0 and 1) for the rest instances
		else
		{
			if (displayMode == 3) printf("%.3f rest from %.3f\n", Distance[instanceCounter],bestValue);
			for (mCounter = 1; mCounter <= TotalMitigations; mCounter++)
			{
				//keep track of the mitigation's counts if it is not already fixed
				if (FixedMitigations[mCounter] == -1)
				{
					if (Data[instanceCounter][mCounter+2] == 0)
						tempRestFreqCount[mCounter][0]++;
					else if (Data[instanceCounter][mCounter+2] == 1)
						tempRestFreqCount[mCounter][1]++;
				}
			}
		}
	}

	float maxScore = -infinity;
	int maxScoredMitigation = 0;
	float maxScoredMitigationStatus = -1;
	float best,rest;

	//normalize each frequency count by dividing it by the total number of instances and score each mitigation using the best^2/(best+rest) and keep min and max
	for (mCounter = 1; mCounter <= TotalMitigations; mCounter++)
	{
		//do this only if mitigation is not fixed already
		if (FixedMitigations[mCounter] == -1)
		{
			//find the score of the mitigation when it is off
			best = tempBestFreqCount[mCounter][0]/TotalInstances;
			rest = tempRestFreqCount[mCounter][0]/TotalInstances;

			if (best == 0 && rest == 0)
				tempScoreOff[mCounter] = 0;
			else
				tempScoreOff[mCounter] = pow(best,2)/(best+rest);

			if (displayMode == 3) printf( "m%d with best:%.3f and rest:%.3f\n",mCounter,best,rest);

			//keep its information if it is the max score seen so far
			if (tempScoreOff[mCounter] > maxScore)
			{
				maxScore = tempScoreOff[mCounter];
				maxScoredMitigation = mCounter;
				maxScoredMitigationStatus = 0;
			}

			//find the score of the mitigation when it is on
			best = tempBestFreqCount[mCounter][1]/TotalInstances;
			rest = tempRestFreqCount[mCounter][1]/TotalInstances;

			if (best == 0 && rest == 0)
				tempScoreOn[mCounter] = 0;
			else
				tempScoreOn[mCounter] = pow(best,2)/(best+rest);

			if (displayMode == 3) printf( "m%d with best:%.3f and rest:%.3f\n",mCounter,best,rest);

			//keep its information if it is the max score seen so far
			if (tempScoreOn[mCounter] > maxScore)
			{
				maxScore = tempScoreOn[mCounter];
				maxScoredMitigation = mCounter;
				maxScoredMitigationStatus = 1;
			}
			if (displayMode == 3) printf( "score of m%d 0:%.3f 1:%.3f\n",mCounter,tempScoreOff[mCounter],tempScoreOn[mCounter]);
		}
	}

	if (displayMode == 3) printf( "chosen mitigation is m%d with status %.0f has score %.3f\n",maxScoredMitigation,maxScoredMitigationStatus,maxScore);

	FixedMitigations[maxScoredMitigation] = maxScoredMitigationStatus;
	recentSetMitigation = maxScoredMitigation;
	recentSetMitigationStatus = maxScoredMitigationStatus;
}

void sweetSpot()
{
	if (displayMode == 3) printf("MIN and MAX %.3f,%.3f,%.3f,%.3f\n",MinCost,MaxCost,MinAtt,MaxAtt);
	float normalizedCost,normalizedAtt;
	//normalize the att and cost using their
	for (instanceCounter = 1; instanceCounter <= TotalInstances; instanceCounter++)
	{
		//calculate the normalized cost and att (if max and min are the same, the small numbers added allow the distance to be correct)
		normalizedCost = (Data[instanceCounter][1] - MinCost)/(MaxCost - MinCost + small);
		normalizedAtt = (Data[instanceCounter][2] - MinAtt + small)/(MaxAtt - MinAtt + small);

		//distance of each instance is measured from normalized (cost=0,att=1) which is the sweet spot
		Distance[instanceCounter] = pow( pow((normalizedCost - 0),2) + pow((normalizedAtt - 1),2 ),0.5);
	}
}

int selectValue(int val1, int val2)
{
	double randomValue = (double)rand()/((double)(RAND_MAX)+(double)(1));
	int returnValue;

	if (randomValue < 0.5)
		returnValue = val1;
	else
		returnValue = val2;
	return returnValue;
}

void addInstance(float costVar, float attVar)
{
	TotalInstances++;
	Data[TotalInstances][1] = costVar;
	Data[TotalInstances][2] = attVar;

	for (mCounter = 1; mCounter <= TotalMitigations; mCounter++)
		Data[TotalInstances][mCounter+2] = mArray[mCounter];

	if (MinCost > Data[TotalInstances][1])
		MinCost = Data[TotalInstances][1];
	if (MaxCost < Data[TotalInstances][1])
		MaxCost = Data[TotalInstances][1];

	if (MinAtt > Data[TotalInstances][2])
		MinAtt = Data[TotalInstances][2];
	if (MaxAtt < Data[TotalInstances][2])
		MaxAtt = Data[TotalInstances][2];
}

float minValue(float val1, float val2)
{
	if (val1 < val2)
		return val1;
	else
		return val2;
}
